import simplepbr
from direct.showbase.ShowBase import ShowBase
from panda3d.core import DirectionalLight, AmbientLight, CardMaker
from panda3d.core import Vec3, Vec4, MaterialAttrib
from fix_gltf_bam import render_stage_convert

# This is a sample program to demonstrate various types of
# texture and material fixes needed to get models working
# with a .blend->gltf->bam workflow, with shadows and using simplepbr.
# This uses Panda3d default "mouse movement" so the camera is initially
# loaded at the world origin. Use the mouse to browse around from different angles.
#
# This demo lays out two models in each of four cases:
# 1) Direct load of .blend
# 2) Direct load of .gltf exported from the same .blend
# 3) #1 with fixes applied (see flags below)
# 4) #2 with fixes applied
#
# The two models are both 1x1m cubes with:
# 1) Metallic blue PBR material (loaded closer to origin)
# 2) A grey/blue "building" texture
#
# Plus a simple texture-loaded card acting as a floor.
# In all cases the generated card with a simple loaded texture
# renders properly with directional lighting applied as expected.
#
# The light is set up at (-10, -10, 10) to cast shadows diagonally
# into the distance.

#
# Types of fixes to apply to models (see links below)
# #1 - This works around https://github.com/Moguri/blend2bam/issues/78
#
do_fix_material = True

# #2 - This is a workaround for simplepbr handling of gltf texture inputs
# https://discourse.panda3d.org/t/directional-lighting-not-working-when-imported-with-panda3d-gltf-but-does-when-imported-with-assimp/29524
do_fix_render_stage = True

# fix_material_gltf() gets the glTF textures working in the
# gltf pipeline, otherwise they are not rendered at all,
# and they are missing when importing directly from .blend.
#
# However, with this fix alone, PBR materials don't get directional lighting,
# which by my understanding from the link above is due to how PBR texture
# inputs are generated in the .bam file.
#
# fix_render_stage gets directional lighting to work, however
# it seems to break the textures that were working when using fix_material_gltf alone.


def fix_material_gltf(node):
    for node_path in node.find_all_matches('**/+GeomNode'):
        geom_states = node_path.node().get_geom_states()
        for geom_state in geom_states:
            if geom_state.has_attrib(MaterialAttrib):
                mat = geom_state.get_attrib(MaterialAttrib).get_material()
                print(f"Fixed material for {node}")
                mat.clear_emission()


if __name__ == "__main__":
    base = ShowBase()
    simplepbr.init()

    # Parent nodes by model type
    # Blender: left side
    blendernodes = base.render.attachNewNode("blender")
    blendernodes.setPos(0, 0, 0)

    # glTF: right side
    gltfnodes = base.render.attachNewNode("gltf")
    gltfnodes.setPos(10, 0, 0)

    # Parents by fixed (front)/unfixed (back)
    blenderfixed = blendernodes.attachNewNode("bfixed")
    blenderfixed.setPos(0, 0, 0)
    blenderunfixed = blendernodes.attachNewNode("bunfixed")
    blenderunfixed.setPos(0, 10, 0)
    gltffixed = gltfnodes.attachNewNode("gfixed")
    gltffixed.setPos(0, 0, 0)
    gltfunfixed = gltfnodes.attachNewNode("gunfixed")
    gltfunfixed.setPos(0, 10, 0)

    loadedmodels = []

    # Load and place all the models
    for models, nodes in [
                # Blender files
                (["models/cube.blend", "models/1x1base.blend"],
                 [blenderfixed, blenderunfixed]),

                # glTF files
                (["models/cube.gltf", "models/1x1base.gltf"],
                 [gltffixed, gltfunfixed])
            ]:
        for n, model in enumerate(models):
            for parent in nodes:
                node = base.loader.loadModel(model)
                loadedmodels.append(node)
                node.reparentTo(parent)
                # Offset successive models diagonally
                idx = n + 1
                node.setPos(idx*2, idx*2, 0)

    # Apply fixes to the fixed nodes
    if do_fix_render_stage:
        render_stage_convert(blenderfixed)
        render_stage_convert(gltffixed)

    if do_fix_material:
        fix_material_gltf(blenderfixed)
        fix_material_gltf(gltffixed)

    # Create ground plane
    ground = CardMaker("ground")
    width = height = 20
    ground.setFrame(  # ll, lr, ur, ul (counterclockwise)
        Vec3(0, 0, 0), Vec3(width, 0, 0),
        Vec3(width, height, 0), Vec3(0, height, 0))
    groundnp = base.render.attach_new_node(ground.generate())
    groundnp.setTexture(base.loader.loadTexture("models/watertex.png"))

    # Create colored directional on left behind origin
    light = DirectionalLight("light")
    light.showFrustum()
    light.setColor(Vec4(0.8, 0, 1.0, 1))
    lightnp = base.render.attach_new_node(light)
    lightnp.reparentTo(base.camera)
    lightnp.setPos(-10, -10, 10)
    lightnp.look_at(0, 0, 0)
    base.render.set_light(lightnp)

    # Enable shadows
    light.setShadowCaster(True, 1024, 1024)
    light.setScene(base.render)
    light.getLens().setFilmSize(512, 512)
    light.getLens().setFov(45)
    light.getLens().setNearFar(2, 100)
    base.render.setShaderAuto()

    base.run()
